%%% @doc
%%% Clutch Controller
%%%
%%% This process is a in charge of maintaining the program's state.
%%% @end

-module(gmc_con).
-vsn("0.1.0").
-author("Craig Everett <craigeverett@qpq.swiss>").
-copyright("Craig Everett <craigeverett@qpq.swiss>").
-license("GPL-3.0-or-later").

-behavior(gen_server).
-export([make_key/6, recover_key/1, mnemonic/1, rename_key/2, drop_key/1,
         login/1, password/2]). % add_key/1, drop_key/1]).
-export([encrypt/2, decrypt/2]).
-export([start_link/0, stop/0, save/1]).
-export([init/1, terminate/2, code_change/3,
         handle_call/3, handle_cast/2, handle_info/2]).
-include("$zx_include/zx_logger.hrl").

-include("gmc.hrl").


%%% Type and Record Definitions


-record(s,
        {window     = none :: none | wx:wx_object(),
         accounts   = []   :: [clutch:ak()],
         pass       = none :: none | binary(),
         keys       = []   :: [clutch:key()],
         chains     = []   :: [chain()],
         prefs      = #{}  :: #{atom() := term()}}).

-record(chain,
        {id    = "" :: string(),
         nodes = [] :: [address()],
         mdws  = [] :: [address()]}).

-type state()   :: #s{}.
-type chain()   :: tuple().
-type address() :: {inet:ip_address(), inet:port_number()}.



%% Interface

-spec make_key(Type, Size, Name, Seed, Encoding, Transform) -> ok
    when Type      :: {eddsa, ed25519},
         Size      :: 256,
         Name      :: string(),
         Seed      :: string(),
         Encoding  :: utf8 | base64 | base58,
         Transform :: raw | {Algo, Yugeness},
         Algo      :: sha3 | sha2 | x_or | pbkdf2,
         Yugeness  :: rand | non_neg_integer().
%% @doc
%% Generate a new key.
%% The magic first two args are here because `ak_*' basically has no option other than
%% to mean a 256 bit key based on Curve 25519. When more key varieties are added to the
%% system this will change quite a lot.

make_key({eddsa, ed25519}, 256, Name, Seed, Encoding, Transform) ->
    gen_server:cast(?MODULE, {make_key, Name, Seed, Encoding, Transform}).


-spec recover_key(Mnemonic) -> ok
    when Mnemonic :: string().

recover_key(Mnemonic) ->
    gen_server:cast(?MODULE, {recover_key, Mnemonic}).


-spec mnemonic(ID) -> {ok, Mnemonic} | error
    when ID       :: clutch:id(),
         Mnemonic :: string().

mnemonic(ID) ->
    gen_server:call(?MODULE, {mnemonic, ID}).


-spec rename_key(ID, NewName) -> ok
    when ID      :: clutch:id(),
         NewName :: string().

rename_key(ID, NewName) ->
    gen_server:cast(?MODULE, {rename_key, ID, NewName}).


-spec drop_key(ID) -> ok
    when ID :: clutch:id().

drop_key(ID) ->
    gen_server:cast(?MODULE, {drop_key, ID}).


-spec stop() -> ok.

stop() ->
    gen_server:cast(?MODULE, stop).


-spec login(Phrase) -> ok
    when Phrase :: string().
    
login(Phrase) ->
    gen_server:cast(?MODULE, {login, Phrase}).


-spec password(Old, New) -> ok
    when Old :: none | string(),
         New :: none | string().

password(Old, New) ->
    gen_server:cast(?MODULE, {password, Old, New}).


-spec save(Prefs) -> ok | {error, Reason}
    when Prefs  :: #{atom() := term()},
         Reason :: file:posix().

save(Prefs) ->
    gen_server:call(?MODULE, {save, Prefs}).



%%% Startup Functions


-spec start_link() -> Result
    when Result :: {ok, pid()}
                 | {error, Reason},
         Reason :: {already_started, pid()}
                 | {shutdown, term()}
                 | term().
%% @private
%% Called by gmc_sup.

start_link() ->
    gen_server:start_link({local, ?MODULE}, ?MODULE, none, []).


-spec init(none) -> {ok, state()}.

init(none) ->
    ok = log(info, "Starting"),
    Prefs = read_prefs(),
    Window = gmc_gui:start_link(Prefs),
    ok = log(info, "Window: ~p", [Window]),
    ok = gmc_gui:ask_password(),
    State = #s{window = Window},
    ArgV = zx_daemon:argv(),
    ok = gmc_gui:show(ArgV),
    {ok, State}.


read_prefs() ->
    case file:consult(prefs_path()) of
        {ok, Prefs} ->
            proplists:to_map(Prefs);
        _ ->
            #{selected => 0,
              lang     => en_us,
              geometry => none}
    end.



%%% gen_server Message Handling Callbacks


-spec handle_call(Message, From, State) -> Result
    when Message  :: term(),
         From     :: {pid(), reference()},
         State    :: state(),
         Result   :: {reply, Response, NewState}
                   | {noreply, State},
         Response :: ok
                   | {error, {listening, inet:port_number()}},
         NewState :: state().
%% @private
%% The gen_server:handle_call/3 callback.
%% See: http://erlang.org/doc/man/gen_server.html#Module:handle_call-3

handle_call({save, Prefs}, _, State) ->
    Response = do_save(State#s{prefs = Prefs}),
    {reply, Response, State};
handle_call({mnemonic, ID}, _, State) ->
    Response = do_mnemonic(ID, State),
    {reply, Response, State};
handle_call(Unexpected, From, State) ->
    ok = log(warning, "Unexpected call from ~tp: ~tp~n", [From, Unexpected]),
    {noreply, State}.


-spec handle_cast(Message, State) -> {noreply, NewState}
    when Message  :: term(),
         State    :: state(),
         NewState :: state().
%% @private
%% The gen_server:handle_cast/2 callback.
%% See: http://erlang.org/doc/man/gen_server.html#Module:handle_cast-2

handle_cast({make_key, Name, Seed, Encoding, Transform}, State) ->
    NewState = do_make_key(Name, Seed, Encoding, Transform, State),
    {noreply, NewState};
handle_cast({recover_key, Mnemonic}, State) ->
    NewState = do_recover_key(Mnemonic, State),
    {noreply, NewState};
handle_cast({rename_key, ID, NewName}, State) ->
    NewState = do_rename_key(ID, NewName, State),
    {noreply, NewState};
handle_cast({drop_key, ID}, State) ->
    NewState = do_drop_key(ID, State),
    {noreply, NewState};
handle_cast({login, Phrase}, State) ->
    NewState = do_login(Phrase, State),
    {noreply, NewState};
handle_cast({password, Old, New}, State) ->
    NewState = do_password(Old, New, State),
    {noreply, NewState};
handle_cast(stop, State) ->
    ok = log(info, "Received a 'stop' message."),
    {stop, normal, State};
handle_cast(Unexpected, State) ->
    ok = tell(warning, "Unexpected cast: ~tp~n", [Unexpected]),
    {noreply, State}.


-spec handle_info(Message, State) -> {noreply, NewState}
    when Message  :: term(),
         State    :: state(),
         NewState :: state().
%% @private
%% The gen_server:handle_info/2 callback.
%% See: http://erlang.org/doc/man/gen_server.html#Module:handle_info-2

handle_info(Unexpected, State) ->
    ok = log(warning, "Unexpected info: ~tp~n", [Unexpected]),
    {noreply, State}.



%% @private
%% gen_server callback to handle state transformations necessary for hot
%% code updates. This template performs no transformation.

code_change(_, State, _) ->
    {ok, State}.


terminate(Reason, State) ->
    ok = log(info, "Reason: ~tp, State: ~tp", [Reason, State]),
    zx:stop().



%%%

do_make_key(Name, <<>>, _, Transform, State) ->
    Bin = crypto:strong_rand_bytes(32),
    do_make_key2(Name, Bin, Transform, State);
do_make_key(Name, Seed, utf8, Transform, State) ->
    Bin = unicode:characters_to_binary(Seed),
    do_make_key2(Name, Bin, Transform, State);
do_make_key(Name, Seed, base64, Transform, State) ->
    case base64_decode(Seed) of
        {ok, Bin} ->
            do_make_key2(Name, Bin, Transform, State);
        {error, Reason} ->
            ok = gmc_gui:notify({error, {base64, Reason}}),
            State
    end;
do_make_key(Name, Seed, base58, Transform, State) ->
    case base58:check_base58(Seed) of
        true  ->
            Bin = base58:base58_to_binary(Seed),
            do_make_key2(Name, Bin, Transform, State);
        false ->
            ok = gmc_gui:notify({error, {base58, badarg}}),
            State
    end.

do_make_key2(Name, Bin, Transform, State = #s{accounts = Accounts, keys = Keys}) ->
    T = transform(Transform),
    Seed = T(Bin),
    Key = #key{name = KeyName, id = ID} = gmc_key_master:make_key(Name, Seed),
    Account = #ak{name = KeyName, id = ID},
    NewKeys = [Key | Keys],
    NewAccounts = [Account | Accounts],
    ok = gmc_gui:show(NewAccounts),
    State#s{accounts = NewAccounts, keys = NewKeys}.


base64_decode(String) ->
    try
        {ok, base64:decode(String)}
    catch
        E:R -> {E, R}
    end.


transform({sha3, 256}) ->
    fun(D) -> crypto:hash(sha3_256, D) end;
transform({sha2, 256}) ->
    fun(D) -> crypto:hash(sha256, D) end;
transform({x_or, 256}) ->
    fun t_xor/1;
transform(raw) ->
    fun(D) -> D end.


t_xor(Bin) -> t_xor(Bin, <<0:256>>).

t_xor(<<H:32/binary, T/binary>>, A) ->
    t_xor(T, crypto:exor(H, A));
t_xor(<<>>, A) ->
    A;
t_xor(B, A) ->
    H = <<0:(256 - bit_size(B)), B/binary>>,
    crypto:exor(H, A).


do_recover_key(Mnemonic, State = #s{keys = Keys, accounts = Accounts}) ->
    case gmc_key_master:decode(Mnemonic) of
        {ok, Seed} ->
            do_recover_key2(Seed, State);
        Error ->
            ok = gmc_gui:notify(Error),
            State
    end.

do_recover_key2(Seed, State = #s{keys = Keys, accounts = Accounts}) ->
    Recovered = #key{id = ID, name = Name} = gmc_key_master:make_key("", Seed),
    case lists:keymember(ID, #key.id, Keys) of
        false ->
            NewKeys = [Recovered | Keys],
            Account = #ak{name = Name, id = ID},
            NewAccounts = [Account | Accounts],
            ok = gmc_gui:show(NewAccounts),
            State#s{accounts = NewAccounts, keys = NewKeys};
        true ->
            State
    end.


do_mnemonic(ID, #s{keys = Keys}) ->
    case lists:keyfind(ID, #key.id, Keys) of
        #key{pair = #{secret := <<K:32/binary, _/binary>>}} ->
            Mnemonic = gmc_key_master:encode(K),
            {ok, Mnemonic};
        false ->
            {error, bad_key}
    end.


do_rename_key(ID, NewName, State = #s{accounts = Accounts, keys = Keys}) ->
    A = lists:keyfind(ID, #ak.id, Accounts),
    K = lists:keyfind(ID, #key.id, Keys),
    NewAccounts = lists:keystore(ID, #ak.id, Accounts, A#ak{name = NewName}),
    NewKeys = lists:keystore(ID, #key.id, Keys, K#key{name = NewName}),
    ok = gmc_gui:show(NewAccounts),
    State#s{accounts = NewAccounts, keys = NewKeys}.


do_drop_key(ID, State = #s{accounts = Accounts, keys = Keys}) ->
    NewAccounts = lists:keydelete(ID, #ak.id, Accounts),
    NewKeys = lists:keydelete(ID, #key.id, Keys),
    ok = gmc_gui:show(NewAccounts),
    State#s{accounts = NewAccounts, keys = NewKeys}.


encrypt(Pass, Binary) ->
    Flags = [{encrypt, true}, {padding, pkcs_padding}],
    crypto:crypto_one_time(aes_256_ecb, Pass, Binary, Flags).


decrypt(Pass, Binary) ->
    Flags = [{encrypt, false}, {padding, pkcs_padding}],
    crypto:crypto_one_time(aes_256_ecb, Pass, Binary, Flags).


pass(Phrase) ->
    crypto:hash(sha3_256, Phrase).


do_login(none, State) ->
    case read(none) of
        {ok, Recovered = #s{accounts = Accounts}} ->
            ok = gmc_gui:show(Accounts),
            Recovered;
        error ->
            State
    end; 
do_login(Phrase, State = #s{pass = none}) ->
    Pass = pass(Phrase),
    try
        case read(Pass) of
            {ok, Recovered = #s{accounts = Accounts}} ->
                ok = gmc_gui:show(Accounts),
                Recovered;
            error ->
                State#s{pass = Pass}
        end
    catch
        E:R ->
            ok = tell("Problem with loading state: {~p, ~p}", [E, R]),
            ok = gmc_gui:ask_password(),
            State
    end.


do_password(none, none, State) ->
    State;
do_password(none, New, State = #s{pass = none}) ->
    Pass = pass(New),
    State#s{pass = Pass};
do_password(Old, none, State = #s{pass = Pass}) ->
    case pass(Old) =:= Pass of
        true  -> State#s{pass = none};
        false -> State
    end;
do_password(Old, New, State = #s{pass = Pass}) ->
    case pass(Old) =:= Pass of
        true  -> State#s{pass = pass(New)};
        false -> State
    end.


do_save(State = #s{prefs = Prefs}) ->
    ok = persist(Prefs),
    do_save2(State).


do_save2(State = #s{pass = none}) ->
    Path = save_path(),
    ok = filelib:ensure_dir(Path),
    file:write_file(Path, term_to_binary(State));
do_save2(State = #s{pass = Pass}) ->
    Path = save_path(),
    ok = filelib:ensure_dir(Path),
    Cipher = encrypt(Pass, term_to_binary(State)),
    file:write_file(Path, Cipher).


read(none) ->
    case file:read_file(save_path()) of
        {ok, Bin}       -> {ok, binary_to_term(Bin)};
        {error, enoent} -> error
    end;
read(Pass) ->
    case file:read_file(save_path()) of
        {ok, Cipher}    -> {ok, binary_to_term(decrypt(Pass, Cipher))};
        {error, enoent} -> error
    end.


save_path() ->
    filename:join(zx_lib:path(var, "otpr", "clutch"), "opaque.data").


persist(Prefs) ->
    Path = prefs_path(),
    ok = filelib:ensure_dir(Path),
    zx_lib:write_terms(Path, proplists:from_map(Prefs)).


prefs_path() ->
    filename:join(zx_lib:path(etc, "otpr", "clutch"), "prefs.eterms").
